
import pickle
import pandas as pd
import lime
import lime.lime_tabular
from .train import check_input_format, preprocess_dataset



def adapt_explanations(explanations: list, 
                      feature_names: list):
    print("[ML] Adapting the explanations")
    
    explainable_predictions = []
    for explanation in explanations:
        # Mapper from id to feature_name
        id_to_feature_name = {i: name for i, name in enumerate(feature_names)}

        # Recover feature contributions
        contributions_with_ids = explanation.as_map()[1]

        # Transform id in to feature name
        contributions = [(id_to_feature_name[id], contrib) for id, contrib in contributions_with_ids]

        # Prepare output dictionary
        explainable_predictions.append({
            'explainability_tool': 'lime',
            'explanation_type': 'simplification, feature_relevance',
            'prediction': explanation.local_pred[0],
            'contributions': contributions,
            'intercept': explanation.intercept[0],
            'predicted_value': explanation.predicted_value,
            'score': explanation.score
        })
    
    return explainable_predictions


def explain(
    input_data: pd.DataFrame,
    model_path: str,
    preprocess: bool = False,
):
    """
    Inputs:
    ------
    input_data: Data included as a Pandas DataFrame, whose first columsn are the targets,
                and subsequent columns are the independent values.
    model_path: The path that the pipeline will use to load the trained model
    preprocess (optional): If the input_data is not preprocessed, apply the modifications
        
    Outputs:
    -------
    trained_model: To be passed to the explainer
    input_data: Actually, preprocessed data to be passed to the explainer
    """
    
    try:
        print("[ML] Loading trained model")
        with open(model_path, 'rb') as model_file:
            model = pickle.load(model_file)
        
        """
        Data preparation section (Check the input data is in the proper format and raise an error if not)
        """
        cols = input_data.columns
        y_column = cols[:1].to_list()
        X_columns = cols[1:].to_list()
        check_input_format(input_data, X_columns, y_column)

        if(preprocess):
            input_data = preprocess_dataset(input_data)
        else:
            print("[ML] Input data NOT preprocessed, it was done before")

        """
        Explain instance
        """
        explainer = {
            "training_data":input_data[X_columns].to_numpy(),
            "feature_names": X_columns,
            "verbose":True,
            "mode":'regression',
        }
        explainer = lime.lime_tabular.LimeTabularExplainer(**explainer)
        
        print("[ML] Performing the explanations")
        
        explanations = []
        for prediction in input_data[X_columns].to_numpy():
            explanations.append(
                explainer.explain_instance(
                    prediction,
                    model.predict
                )
            )

        """
        Prepare explanation to output format
        """
        explanations = adapt_explanations(explanations, model.X_columns)
        
    except AssertionError as err:
        raise ValueError("Explainer Error: Input data is not in the proper format") from err
            
    return {"input_df":input_data, "explanations":explanations}
        
#
