import pickle
from datetime import datetime
import pandas as pd
from .train import check_input_format, preprocess_dataset


def inference(
    input_data: pd.DataFrame,
    model_path: str,
    preprocess: bool = False,
):
    """
    Inputs:
    ------
    input_data: Data included as a Pandas DataFrame, whose first columsn are the targets,
                and subsequent columns are the independent values.
    model_path: The path that the pipeline will use to load the trained model
    preprocess (optional): If the input_data is not preprocessed, apply the modifications
                    
    Outputs:
    -------
    trained_model: To be passed to the explainer
    input_data: Actually, preprocessed data to be passed to the explainer
    """

    try:
        with open(model_path, 'rb') as model_file:
            model = pickle.load(model_file)

        """
        Data preparation section (Check the input data is in the proper format and raise an error if not)
        """
        cols = input_data.columns
        y_column = cols[:1].to_list()
        X_columns = cols[1:].to_list()
        check_input_format(input_data, X_columns, y_column)
        
        if(preprocess):
            input_data = preprocess_dataset(input_data)
        else:
            print("[ML] Input data NOT preprocessed, it was done before")

        """
        Model prediction
        """        
        print("[ML] Performing inference")
        prediction = model.predict(input_data[model.X_columns].to_numpy())
        
        """
        Adaptation of the model prediction to an agreed format
        """
        results = pd.DataFrame.from_dict({
            'timestamp': [datetime.now().strftime("%d-%m-%Y %H:%M:%S")]*len(prediction),
            'prediction': prediction,
            'X':input_data[model.X_columns[0]]
        })
        
        
    except AssertionError as err:
        raise ValueError("Inference Error: Input data is not in the proper format") from err

    return {"input_df":input_data, "results_df":results}
