# Import Section (import the necessary packages for training the model)
from typing import *
import pickle
import typing
import pandas as pd
import numpy as np
from sklearn.model_selection import KFold, RepeatedKFold, cross_val_score
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import GridSearchCV


def check_input_format(input_data: pd.DataFrame, X_columns, y_columns):
	"""Check if the input data format is proper to perform the training by a set of conditions

	Args:
		input_data (pd.DataFrame): Dict that contains training features, regression prediction and the name of the features
	"""
	print("[ML] Asserting the format of the dataframe")
	try:
		assert len(input_data), "Input dataframe is empty"
		assert input_data.columns.isin(X_columns+y_columns).sum() >= len(X_columns) + len(y_columns), f"Missing columns in input dataframe. Required columns are: {X_columns+y_columns} but got {input_data.columns}"
		
		X_sliced_data = input_data[X_columns]
		assert not X_sliced_data.isna().any(axis=None), "There are invalid NAN values in the train values of the input dataframe"
		assert not X_sliced_data.isnull().any(axis=None), "There are invalid null values in the train values of the input dataframe"
		assert not X_sliced_data.dtypes.isin(['string', 'object']).any(axis=None), "Invalid column types in the train values of the input dataframe"
		
		y_sliced_data = input_data[y_columns]
		assert not y_sliced_data.isna().any(axis=None), "There are invalid NAN values in the true values of the input dataframe"
		assert not y_sliced_data.isnull().any(axis=None), "There are invalid null values in the true values of the input dataframe"
		assert not y_sliced_data.dtypes.isin(['string', 'object']).any(axis=None), "Invalid column types in the true values of the input dataframe"
	except AssertionError as err:
		raise ValueError("Training Error: Input data is not in the proper format") from err


def preprocess_dataset(input_data: pd.DataFrame) -> pd.DataFrame:
	"""Optional function to apply specific preprocess to input data
	"""
	print("[ML] Preprocessing the input data")
	for col in input_data.columns:
		if(input_data[col].dtype != float):
			input_data.loc[:, col] = input_data[col].astype(float)

	return input_data


def evaluate(model: RandomForestRegressor, 
			 X: np.ndarray, 
			 y: np.ndarray, 
			 validator: typing.Any,
			 metric: typing.Any):
	print("[ML] Evaluating the model")
	metrics_mapper = {'R2': 'r2', 'MSE': 'neg_mean_squared_error'}
	
	if(type(validator) == str):
		validator = {"name":validator}
	
	
	if validator['name'] == 'k_fold_cv':
		cv_generator = KFold(validator.get('n_splits', 5),
				   shuffle=validator.get('shuffle', False))
	elif validator['name'] == 'repeated_k_fold':
		cv_generator = RepeatedKFold(n_splits=validator.get('n_splits', 5),
						   n_repeats=validator.get('n_repeats', 10))
	else:
		raise ValueError(f"Wrong validator: {validator}")
		
	# Calculate scores
	scores = cross_val_score(model, X, y, scoring=metrics_mapper.get(metric, metric), cv=cv_generator)

	return scores

def instance_model(hyperparameters: dict, use_grid_search: bool):
	print(f"[ML] Instantiating the model{' using grid search' if use_grid_search else ''}")
	
	if(use_grid_search):
		model = RandomForestRegressor()
		
		formatted_hyper = dict()
		for hkey, hval in hyperparameters.items():
			htype = type(hval)
			if(htype == dict):
				hval_type = hval["type"]
				if(hval_type == "range"):
					if(len(hval.keys() & {'min_value', 'max_value', 'step'}) >= 2):
						formatted_hyper[hkey] = list(range(
							hval.get('min_value', hval.get('max_value',0) - hval.get('step',0)),
							hval.get('max_value', hval['min_value'] + hval['step']),
							hval.get('step', 1)
						))
				elif(hval_type == "discrete"):
					discrete_value = hval.get("chosen_values") and [hval.get("default_value")]
					if(discrete_value and len(discrete_value)):
						formatted_hyper[hkey] = discrete_value
				elif(hval_type == "bool"):
					bool_value = hval.get("chosen_values") and [hval.get("value")]
					if(bool_value and len(bool_value)):
						formatted_hyper[hkey] = bool_value 
			elif(htype == list):
				formatted_hyper[hkey] = hval
			elif(htype in {int, float, str}):
				formatted_hyper[hkey] = [hval]
			
		formatted_pretty = '\n       '.join((
			f"{key}: {str(value_list)}"
			if len(value_list) <= 5 else
			f"{key}: [{', '.join(map(str, value_list[:2]))}, ..., {', '.join(map(str, value_list[-2:]))}]"
			for key, value_list in formatted_hyper.items()
		))
		print(f"[ML]  Grid search parameters:\n       {formatted_pretty}")
		
		return GridSearchCV(
			model,
			formatted_hyper
		)
	else:
		return RandomForestRegressor(**hyperparameters)


def train(input_data: pd.DataFrame,
		  model_session_info: dict,
		  model_path: str,):
	"""
	Inputs:
	------
	input_data: Data included as a Pandas DataFrame, whose first columsn are the targets,
                and subsequent columns are the independent values.
	model_session_info: Information related to the training configuration 
						(here it is included both hyperparameters and validation strategy)
	model_path: The path that the pipeline will use to store the newly created model
					
	Outputs:
	-------
	trained_model: To be passed to the explainer
	input_data: Actually, preprocessed data to be passed to the explainer
	"""
	
	model = None
	model_session_info = model_session_info['configuration']

	cols = input_data.columns

	y_column = cols[:1].to_list()
	X_columns = cols[1:].to_list()

	# Retrieve hyperparameters and validation strategy from
	hyperparameters = model_session_info['hyperparameters']
	use_grid_search = model_session_info['grid_search']

	check_input_format(input_data, X_columns, y_column)
	input_data = preprocess_dataset(input_data)

	"""
	Model instantiation & training
	"""
	model = instance_model(hyperparameters, use_grid_search)

	print("[ML] Training the model")
	#with mlflow_run:
	model = model.fit(input_data[X_columns].to_numpy(), np.squeeze(input_data[y_column].to_numpy()))

	model.X_columns = X_columns
	model.y_column = y_column
	
	evaluation_scores = input_data

	"""
	Model storage
	"""
	print("[ML] Storing the model")
	with open(model_path, 'wb+') as model_file:
		pickle.dump(model, model_file)

	return {
	    "input_data": input_data, 
	    "evaluation_metrics": evaluation_scores,
	}
			

#
